#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains project version information.

.. currentmodule:: ghostbusterV4.version
.. moduleauthor:: shubham_shah <sshah@assetnote.io>
"""

__version__ = "1.0.5"  #: the working version
__release__ = "1.0.5"  #: the release version
